<link href='https://fonts.googleapis.com/css?family=Oxygen:300,400,700' rel='stylesheet'>
<script async defer src="https://buttons.github.io/buttons.js"></script>

{% if prerelease %}
<script>
$(document).ready(function() {
    $('div.body').prepend('<div id="dev-warning">Warning: This is the documentation for a pre-release version of Falcon.</div>')
});
</script>
{% endif %}

<div id="logo">
    <a href="{{ pathto(master_doc) }}"><img src="{{ pathto('_static/img/logo.svg', 1) }}" width="163" height="163" alt="Falcon Web Framework Logo"/></a>
</div>

<div id="gh-buttons">
    <a class="github-button" href="https://github.com/falconry/falcon" data-icon="octicon-star" aria-label="Star falconry/falcon on GitHub" data-size="large">Star</a>
    <a class="github-button" href="https://github.com/falconry/falcon/issues" data-icon="octicon-issue-opened" aria-label="Issue falconry/falcon on GitHub" data-size="large">Issue</a>
</div>
